@@ -2,32 +2,9 @@
     commands/changeexpirycommand.cpp
 
     This file is part of Kleopatra, the KDE keymanager
-    Copyright (c) 2008 Klarälvdalens Datakonsult AB
+    SPDX-FileCopyrightText: 2008 Klarälvdalens Datakonsult AB
 
-    Kleopatra is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    Kleopatra is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
-
-    In addition, as a special exception, the copyright holders give
-    permission to link the code of this program with any edition of
-    the Qt library by Trolltech AS, Norway (or with modified versions
-    of Qt that use the same license as Qt), and distribute linked
-    combinations including the two.  You must obey the GNU General
-    Public License in all respects for all of the code used other than
-    Qt.  If you modify this file, you may extend this exception to
-    your version of the file, but you are not obligated to do so.  If
-    you do not wish to do so, delete this exception statement from
-    your version.
+    SPDX-License-Identifier: GPL-2.0-or-later
 */
 
 #include <config-kleopatra.h>
@@ -50,6 +27,12 @@
 
 #include <QDateTime>
 
+#include <QDebug>
+
+#include <gpgme++/gpgmepp_version.h>
+#if GPGMEPP_VERSION >= 0x10E01 // 1.14.1
+# define CHANGEEXPIRYJOB_SUPPORTS_SUBKEYS
+#endif
 
 using namespace Kleo;
 using namespace Kleo::Commands;
@@ -83,6 +66,7 @@ private:
 
 private:
     GpgME::Key key;
+    GpgME::Subkey subkey;
     QPointer<ExpiryDialog> dialog;
     QPointer<ChangeExpiryJob> job;
 };
@@ -141,9 +125,13 @@ ChangeExpiryCommand::~ChangeExpiryCommand()
     qCDebug(KLEOPATRA_LOG);
 }
 
+void ChangeExpiryCommand::setSubkey(const GpgME::Subkey &subkey)
+{
+    d->subkey = subkey;
+}
+
 void ChangeExpiryCommand::doStart()
 {
-
     const std::vector<Key> keys = d->keys();
     if (keys.size() != 1 ||
             keys.front().protocol() != GpgME::OpenPGP ||
@@ -155,10 +143,20 @@ void ChangeExpiryCommand::doStart()
 
     d->key = keys.front();
 
+    if (!d->subkey.isNull() &&
+            d->subkey.parent().primaryFingerprint() != d->key.primaryFingerprint()) {
+        qDebug() << "Invalid subkey" << d->subkey.fingerprint()
+                 << ": Not a subkey of key" << d->key.primaryFingerprint();
+        d->finished();
+        return;
+    }
+
+    const Subkey subkey = !d->subkey.isNull() ? d->subkey : d->key.subkey(0);
+
     d->ensureDialogCreated();
     Q_ASSERT(d->dialog);
-    const Subkey subkey = d->key.subkey(0);
-    d->dialog->setDateOfExpiry(subkey.neverExpires() ? QDate() : QDateTime::fromSecsSinceEpoch(d->key.subkey(0).expirationTime()).date());
+    d->dialog->setDateOfExpiry(subkey.neverExpires() ? QDate() :
+                               QDateTime::fromSecsSinceEpoch(subkey.expirationTime()).date());
     d->dialog->show();
 
 }
@@ -167,7 +165,7 @@ void ChangeExpiryCommand::Private::slotDialogAccepted()
 {
     Q_ASSERT(dialog);
 
-    static const QTime END_OF_DAY(23, 59, 59);   // not used, so as good as any
+    static const QTime END_OF_DAY(23, 59, 59);
 
     const QDateTime expiry(dialog->dateOfExpiry(), END_OF_DAY);
 
@@ -176,7 +174,16 @@ void ChangeExpiryCommand::Private::slotDialogAccepted()
     createJob();
     Q_ASSERT(job);
 
+    std::vector<Subkey> subkeys;
+    if (!subkey.isNull()) {
+        subkeys.push_back(subkey);
+    }
+
+#ifdef CHANGEEXPIRYJOB_SUPPORTS_SUBKEYS
+    if (const Error err = job->start(key, expiry, subkeys)) {
+#else
     if (const Error err = job->start(key, expiry)) {
+#endif
         showErrorDialog(err);
         finished();
     }
