@@ -236,13 +236,13 @@ void KeyToCardCommand::Private::startKeyToOpenPGPCard() {
     std::string existingKey;
     QString encKeyWarning;
     if (slot == 1) {
-        existingKey = pgpCard->sigFpr();
+        existingKey = pgpCard->keyFingerprint(OpenPGPCard::pgpSigKeyRef());
     } else if (slot == 2) {
-        existingKey = pgpCard->encFpr();
+        existingKey = pgpCard->keyFingerprint(OpenPGPCard::pgpEncKeyRef());
         encKeyWarning = i18n("It will no longer be possible to decrypt past communication "
                                 "encrypted for the existing key.");
     } else if (slot == 3) {
-        existingKey = pgpCard->authFpr();
+        existingKey = pgpCard->keyFingerprint(OpenPGPCard::pgpAuthKeyRef());
     }
     if (!existingKey.empty()) {
         const QString message = i18nc("@info",
@@ -364,7 +364,7 @@ void KeyToCardCommand::Private::startKeyToPIVCard()
 
     // Check if we need to do the overwrite warning.
     if (!overwriteExistingAlreadyApproved) {
-        const std::string existingKey = pivCard->keyGrip(cardSlot);
+        const std::string existingKey = pivCard->keyInfo(cardSlot).grip;
         if (!existingKey.empty() && (existingKey != subkey.keyGrip())) {
             const QString decryptionWarning = (cardSlot == PIVCard::keyManagementKeyRef()) ?
                 i18n("It will no longer be possible to decrypt past communication encrypted for the existing key.") :
