@@ -11,9 +11,17 @@
 
 #include "certificateselectiondialog.h"
 
+#include "settings.h"
+
+#include "conf/groupsconfigdialog.h"
+
+#include <view/keytreeview.h>
 #include <view/searchbar.h>
 #include <view/tabwidget.h>
 
+#include "utils/tags.h"
+
+#include <Libkleo/KeyGroup>
 #include <Libkleo/KeyListModel>
 #include <Libkleo/KeyCache>
 
@@ -25,6 +33,7 @@
 #include <gpgme++/key.h>
 
 #include <KLocalizedString>
+#include <KConfigDialog>
 #include <KConfigGroup>
 #include <KSharedConfig>
 
@@ -38,11 +47,26 @@
 
 #include <algorithm>
 
+#include <gpgme++/gpgmepp_version.h>
+#if GPGMEPP_VERSION >= 0x10E00 // 1.14.0
+# define GPGME_HAS_REMARKS
+#endif
+
 using namespace Kleo;
 using namespace Kleo::Dialogs;
 using namespace Kleo::Commands;
 using namespace GpgME;
 
+
+CertificateSelectionDialog::Option CertificateSelectionDialog::optionsFromProtocol(Protocol proto)
+{
+    switch (proto) {
+    case OpenPGP: return CertificateSelectionDialog::OpenPGPFormat;
+    case CMS:     return CertificateSelectionDialog::CMSFormat;
+    default:      return CertificateSelectionDialog::AnyFormat;
+    }
+}
+
 class CertificateSelectionDialog::Private
 {
     friend class ::Kleo::Dialogs::CertificateSelectionDialog;
@@ -72,15 +96,26 @@ private:
         cmd->setParentWidget(q);
         cmd->start();
     }
+    void manageGroups()
+    {
+        KConfigDialog *dialog = KConfigDialog::exists(GroupsConfigDialog::dialogName());
+        if (dialog) {
+            // reparent the dialog to ensure it's shown on top of the modal CertificateSelectionDialog
+            dialog->setParent(q, Qt::Dialog);
+        } else {
+            dialog = new GroupsConfigDialog(q);
+        }
+        dialog->show();
+    }
     void slotKeysMayHaveChanged();
     void slotCurrentViewChanged(QAbstractItemView *newView);
     void slotSelectionChanged();
     void slotDoubleClicked(const QModelIndex &idx);
 
 private:
-    bool acceptable(const std::vector<Key> &keys)
+    bool acceptable(const std::vector<Key> &keys, const std::vector<KeyGroup> &groups)
     {
-        return !keys.empty();
+        return !keys.empty() || !groups.empty();
     }
     void updateLabelText()
     {
@@ -91,81 +126,83 @@ private:
     }
 
 private:
-    QPointer<QAbstractItemView> lastView;
+    QSet<QAbstractItemView *> connectedViews;
     QString customLabelText;
-    Options options;
+    Options options = AnyCertificate | AnyFormat;
 
     struct UI {
         QLabel label;
         SearchBar searchBar;
         TabWidget tabWidget;
         QDialogButtonBox buttonBox;
-        QVBoxLayout vlay;
-
-        explicit UI(CertificateSelectionDialog *q)
-            : label(q),
-              searchBar(q),
-              tabWidget(q),
-              buttonBox(q),
-              vlay(q)
-        {
-            KDAB_SET_OBJECT_NAME(label);
-            KDAB_SET_OBJECT_NAME(searchBar);
-            KDAB_SET_OBJECT_NAME(tabWidget);
-            KDAB_SET_OBJECT_NAME(buttonBox);
-            KDAB_SET_OBJECT_NAME(vlay);
-
-            vlay.addWidget(&label);
-            vlay.addWidget(&searchBar);
-            vlay.addWidget(&tabWidget, 1);
-            vlay.addWidget(&buttonBox);
-
-            QPushButton *const ok = buttonBox.addButton(QDialogButtonBox::Ok);
-            ok->setEnabled(false);
-            QPushButton *const cancel = buttonBox.addButton(QDialogButtonBox::Close);
-            Q_UNUSED(cancel);
-            QPushButton *const reload = buttonBox.addButton(i18n("Reload"),    QDialogButtonBox::ActionRole);
-            QPushButton *const import = buttonBox.addButton(i18n("Import..."),    QDialogButtonBox::ActionRole);
-            QPushButton *const lookup = buttonBox.addButton(i18n("Lookup..."), QDialogButtonBox::ActionRole);
-            QPushButton *const create = buttonBox.addButton(i18n("New..."),    QDialogButtonBox::ActionRole);
-
-            import->setToolTip(i18nc("@info:tooltip", "Import certificate from file"));
-            lookup->setToolTip(i18nc("@info:tooltip", "Lookup certificates on server"));
-            reload->setToolTip(i18nc("@info:tooltip", "Refresh certificate list"));
-            create->setToolTip(i18nc("@info:tooltip", "Create a new certificate"));
-
-            connect(&buttonBox, &QDialogButtonBox::accepted, q, &CertificateSelectionDialog::accept);
-            connect(&buttonBox, &QDialogButtonBox::rejected, q, &CertificateSelectionDialog::reject);
-            connect(reload,     SIGNAL(clicked()),  q, SLOT(reload()));
-            connect(lookup,     SIGNAL(clicked()),  q, SLOT(lookup()));
-            connect(create,     SIGNAL(clicked()),  q, SLOT(create()));
-            connect(KeyCache::instance().get(), SIGNAL(keysMayHaveChanged()),
-                    q, SLOT(slotKeysMayHaveChanged()));
-
-            connect(import, &QPushButton::clicked, q, [import, q] () {
-                import->setEnabled(false);
-                auto cmd = new Kleo::ImportCertificateFromFileCommand();
-                connect(cmd, &Kleo::ImportCertificateFromFileCommand::finished,
-                        q, [import]() {
-                    import->setEnabled(true);
-                });
-                cmd->setParentWidget(q);
-                cmd->start();
-            });
-        }
     } ui;
+
+    void setUpUI(CertificateSelectionDialog *q)
+    {
+        KDAB_SET_OBJECT_NAME(ui.label);
+        KDAB_SET_OBJECT_NAME(ui.searchBar);
+        KDAB_SET_OBJECT_NAME(ui.tabWidget);
+        KDAB_SET_OBJECT_NAME(ui.buttonBox);
+
+        auto *vlay = new QVBoxLayout(q);
+        vlay->addWidget(&ui.label);
+        vlay->addWidget(&ui.searchBar);
+        vlay->addWidget(&ui.tabWidget, 1);
+        vlay->addWidget(&ui.buttonBox);
+
+        QPushButton *const okButton = ui.buttonBox.addButton(QDialogButtonBox::Ok);
+        okButton->setEnabled(false);
+        ui.buttonBox.addButton(QDialogButtonBox::Close);
+        QPushButton *const reloadButton = ui.buttonBox.addButton(i18n("Reload"), QDialogButtonBox::ActionRole);
+        QPushButton *const importButton = ui.buttonBox.addButton(i18n("Import..."), QDialogButtonBox::ActionRole);
+        QPushButton *const lookupButton = ui.buttonBox.addButton(i18n("Lookup..."), QDialogButtonBox::ActionRole);
+        QPushButton *const createButton = ui.buttonBox.addButton(i18n("New..."), QDialogButtonBox::ActionRole);
+        QPushButton *const groupsButton = ui.buttonBox.addButton(i18n("Groups..."), QDialogButtonBox::ActionRole);
+        groupsButton->setVisible(Settings().groupsEnabled());
+
+        importButton->setToolTip(i18nc("@info:tooltip", "Import certificate from file"));
+        lookupButton->setToolTip(i18nc("@info:tooltip", "Lookup certificates on server"));
+        reloadButton->setToolTip(i18nc("@info:tooltip", "Refresh certificate list"));
+        createButton->setToolTip(i18nc("@info:tooltip", "Create a new certificate"));
+        groupsButton->setToolTip(i18nc("@info:tooltip", "Manage certificate groups"));
+
+        connect(&ui.buttonBox, &QDialogButtonBox::accepted, q, &CertificateSelectionDialog::accept);
+        connect(&ui.buttonBox, &QDialogButtonBox::rejected, q, &CertificateSelectionDialog::reject);
+        connect(reloadButton, &QPushButton::clicked, q, [this] () { reload(); });
+        connect(lookupButton, &QPushButton::clicked, q, [this] () { lookup(); });
+        connect(createButton, &QPushButton::clicked, q, [this] () { create(); });
+        connect(groupsButton, &QPushButton::clicked, q, [this] () { manageGroups(); });
+        connect(KeyCache::instance().get(), &KeyCache::keysMayHaveChanged,
+                q, [this] () { slotKeysMayHaveChanged(); });
+
+        connect(importButton, &QPushButton::clicked, q, [importButton, q] () {
+            importButton->setEnabled(false);
+            auto cmd = new Kleo::ImportCertificateFromFileCommand();
+            connect(cmd, &Kleo::ImportCertificateFromFileCommand::finished,
+                    q, [importButton]() {
+                importButton->setEnabled(true);
+            });
+            cmd->setParentWidget(q);
+            cmd->start();
+        });
+    }
 };
 
 CertificateSelectionDialog::Private::Private(CertificateSelectionDialog *qq)
-    : q(qq),
-      ui(q)
+    : q(qq)
 {
+    setUpUI(q);
     ui.tabWidget.setFlatModel(AbstractKeyListModel::createFlatKeyListModel(q));
     ui.tabWidget.setHierarchicalModel(AbstractKeyListModel::createHierarchicalKeyListModel(q));
+#ifdef GPGME_HAS_REMARKS
+    const auto tagKeys = Tags::tagKeys();
+    ui.tabWidget.flatModel()->setRemarkKeys(tagKeys);
+    ui.tabWidget.hierarchicalModel()->setRemarkKeys(tagKeys);
+#endif
     ui.tabWidget.connectSearchBar(&ui.searchBar);
 
-    connect(&ui.tabWidget, SIGNAL(currentViewChanged(QAbstractItemView*)),
-            q, SLOT(slotCurrentViewChanged(QAbstractItemView*)));
+    connect(&ui.tabWidget, &TabWidget::currentViewChanged,
+            q, [this] (QAbstractItemView *view) { slotCurrentViewChanged(view); });
 
     updateLabelText();
     q->setWindowTitle(i18nc("@title:window", "Certificate Selection"));
@@ -199,6 +236,8 @@ QString CertificateSelectionDialog::customLabelText() const
 
 void CertificateSelectionDialog::setOptions(Options options)
 {
+    Q_ASSERT((options & CertificateSelectionDialog::AnyCertificate) != 0);
+    Q_ASSERT((options & CertificateSelectionDialog::AnyFormat) != 0);
     if (d->options == options) {
         return;
     }
@@ -224,21 +263,46 @@ void CertificateSelectionDialog::setKeyFilter(const std::shared_ptr<KeyFilter> &
     d->ui.tabWidget.setKeyFilter(filter);
 }
 
-void CertificateSelectionDialog::selectCertificates(const std::vector<Key> &keys)
+namespace
+{
+
+void selectRows(const QAbstractItemView *view, const QModelIndexList &indexes)
 {
-    const QAbstractItemView *const view = d->ui.tabWidget.currentView();
     if (!view) {
         return;
     }
-    const auto *const model = d->ui.tabWidget.currentModel();
-    Q_ASSERT(model);
     QItemSelectionModel *const sm = view->selectionModel();
     Q_ASSERT(sm);
 
-    Q_FOREACH (const QModelIndex &idx, model->indexes(keys))
+    for (const QModelIndex &idx : qAsConst(indexes)) {
         if (idx.isValid()) {
             sm->select(idx, QItemSelectionModel::Select | QItemSelectionModel::Rows);
         }
+    }
+}
+
+QModelIndexList getGroupIndexes(const KeyListModelInterface *model, const std::vector<KeyGroup> &groups)
+{
+    QModelIndexList indexes;
+    indexes.reserve(groups.size());
+    std::transform(groups.begin(), groups.end(),
+                   std::back_inserter(indexes),
+                   [model] (const KeyGroup &group) {
+                       return model->index(group);
+                   });
+    indexes.erase(std::remove_if(indexes.begin(), indexes.end(),
+                                 [] (const QModelIndex &index) { return !index.isValid(); }),
+                  indexes.end());
+    return indexes;
+}
+
+}
+
+void CertificateSelectionDialog::selectCertificates(const std::vector<Key> &keys)
+{
+    const auto *const model = d->ui.tabWidget.currentModel();
+    Q_ASSERT(model);
+    selectRows(d->ui.tabWidget.currentView(), model->indexes(keys));
 }
 
 void CertificateSelectionDialog::selectCertificate(const Key &key)
@@ -246,17 +310,46 @@ void CertificateSelectionDialog::selectCertificate(const Key &key)
     selectCertificates(std::vector<Key>(1, key));
 }
 
-std::vector<Key> CertificateSelectionDialog::selectedCertificates() const
+void CertificateSelectionDialog::selectGroups(const std::vector<KeyGroup> &groups)
 {
-    const QAbstractItemView *const view = d->ui.tabWidget.currentView();
-    if (!view) {
-        return std::vector<Key>();
-    }
     const auto *const model = d->ui.tabWidget.currentModel();
     Q_ASSERT(model);
+    selectRows(d->ui.tabWidget.currentView(), getGroupIndexes(model, groups));
+}
+
+namespace
+{
+
+QModelIndexList getSelectedRows(const QAbstractItemView *view)
+{
+    if (!view) {
+        return {};
+    }
     const QItemSelectionModel *const sm = view->selectionModel();
     Q_ASSERT(sm);
-    return model->keys(sm->selectedRows());
+    return sm->selectedRows();
+}
+
+std::vector<KeyGroup> getGroups(const KeyListModelInterface *model, const QModelIndexList &indexes)
+{
+    std::vector<KeyGroup> groups;
+    groups.reserve(indexes.size());
+    std::transform(indexes.begin(), indexes.end(),
+                   std::back_inserter(groups),
+                   [model](const QModelIndex &idx) {
+                       return model->group(idx);
+                   });
+    groups.erase(std::remove_if(groups.begin(), groups.end(), std::mem_fn(&Kleo::KeyGroup::isNull)), groups.end());
+    return groups;
+}
+
+}
+
+std::vector<Key> CertificateSelectionDialog::selectedCertificates() const
+{
+    const KeyListModelInterface *const model = d->ui.tabWidget.currentModel();
+    Q_ASSERT(model);
+    return model->keys(getSelectedRows(d->ui.tabWidget.currentView()));
 }
 
 Key CertificateSelectionDialog::selectedCertificate() const
@@ -265,6 +358,13 @@ Key CertificateSelectionDialog::selectedCertificate() const
     return keys.empty() ? Key() : keys.front();
 }
 
+std::vector<KeyGroup> CertificateSelectionDialog::selectedGroups() const
+{
+    const KeyListModelInterface *const model = d->ui.tabWidget.currentModel();
+    Q_ASSERT(model);
+    return getGroups(model, getSelectedRows(d->ui.tabWidget.currentView()));
+}
+
 void CertificateSelectionDialog::hideEvent(QHideEvent *e)
 {
     KSharedConfig::Ptr config = KSharedConfig::openConfig(QStringLiteral("kleopatracertificateselectiondialogrc"));
@@ -279,14 +379,20 @@ void CertificateSelectionDialog::Private::slotKeysMayHaveChanged()
     q->setEnabled(true);
     std::vector<Key> keys = (options & SecretKeys) ? KeyCache::instance()->secretKeys() : KeyCache::instance()->keys();
     q->filterAllowedKeys(keys, options);
-    const std::vector<Key> selected = q->selectedCertificates();
+    const std::vector<KeyGroup> groups = (options & IncludeGroups) ? KeyCache::instance()->groups() : std::vector<KeyGroup>();
+
+    const std::vector<Key> selectedKeys = q->selectedCertificates();
+    const std::vector<KeyGroup> selectedGroups = q->selectedGroups();
     if (AbstractKeyListModel *const model = ui.tabWidget.flatModel()) {
         model->setKeys(keys);
+        model->setGroups(groups);
     }
     if (AbstractKeyListModel *const model = ui.tabWidget.hierarchicalModel()) {
         model->setKeys(keys);
+        model->setGroups(groups);
     }
-    q->selectCertificates(selected);
+    q->selectCertificates(selectedKeys);
+    q->selectGroups(selectedGroups);
 }
 
 void CertificateSelectionDialog::filterAllowedKeys(std::vector<Key> &keys, int options)
@@ -326,20 +432,13 @@ void CertificateSelectionDialog::filterAllowedKeys(std::vector<Key> &keys, int o
 
 void CertificateSelectionDialog::Private::slotCurrentViewChanged(QAbstractItemView *newView)
 {
-    if (lastView) {
-        disconnect(lastView, SIGNAL(doubleClicked(QModelIndex)),
-                   q, SLOT(slotDoubleClicked(QModelIndex)));
-        Q_ASSERT(lastView->selectionModel());
-        disconnect(lastView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)),
-                   q, SLOT(slotSelectionChanged()));
-    }
-    lastView = newView;
-    if (newView) {
-        connect(newView, SIGNAL(doubleClicked(QModelIndex)),
-                q, SLOT(slotDoubleClicked(QModelIndex)));
+    if (!connectedViews.contains(newView)) {
+        connectedViews.insert(newView);
+        connect(newView, &QAbstractItemView::doubleClicked,
+                q, [this] (const QModelIndex &index) { slotDoubleClicked(index); });
         Q_ASSERT(newView->selectionModel());
-        connect(newView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)),
-                q, SLOT(slotSelectionChanged()));
+        connect(newView->selectionModel(), &QItemSelectionModel::selectionChanged,
+                q, [this] (const QItemSelection &, const QItemSelection &) { slotSelectionChanged(); });
     }
     slotSelectionChanged();
 }
@@ -347,7 +446,7 @@ void CertificateSelectionDialog::Private::slotCurrentViewChanged(QAbstractItemVi
 void CertificateSelectionDialog::Private::slotSelectionChanged()
 {
     if (QPushButton *const pb = ui.buttonBox.button(QDialogButtonBox::Ok)) {
-        pb->setEnabled(acceptable(q->selectedCertificates()));
+        pb->setEnabled(acceptable(q->selectedCertificates(), q->selectedGroups()));
     }
 }
 
@@ -357,7 +456,7 @@ void CertificateSelectionDialog::Private::slotDoubleClicked(const QModelIndex &i
     Q_ASSERT(view);
     const auto *const model = ui.tabWidget.currentModel();
     Q_ASSERT(model);
-    Q_UNUSED(model);
+    Q_UNUSED(model)
     QItemSelectionModel *const sm = view->selectionModel();
     Q_ASSERT(sm);
     sm->select(idx, QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
@@ -366,7 +465,7 @@ void CertificateSelectionDialog::Private::slotDoubleClicked(const QModelIndex &i
 
 void CertificateSelectionDialog::accept()
 {
-    if (d->acceptable(selectedCertificates())) {
+    if (d->acceptable(selectedCertificates(), selectedGroups())) {
         QDialog::accept();
     }
 }
