@@ -9,6 +9,8 @@
     SPDX-License-Identifier: GPL-2.0-or-later
 */
 
+#include "keypairinfo.h"
+
 #include <string>
 #include <vector>
 
@@ -18,6 +20,7 @@ namespace Kleo
 {
 namespace SmartCard
 {
+
 /** Class representing an application on a smartcard or similar hardware token. */
 class Card
 {
@@ -73,6 +76,14 @@ public:
 
     QString cardHolder() const;
 
+    void setSigningKeyRef(const std::string &keyRef);
+    std::string signingKeyRef() const;
+    bool hasSigningKey() const;
+
+    void setEncryptionKeyRef(const std::string &keyRef);
+    std::string encryptionKeyRef() const;
+    bool hasEncryptionKey() const;
+
     std::vector<PinState> pinStates() const;
     void setPinStates(const std::vector<PinState> &pinStates);
 
@@ -85,11 +96,18 @@ public:
     QString errorMsg() const;
     void setErrorMsg(const QString &msg);
 
+    const std::vector<KeyPairInfo> & keyInfos() const;
+    const KeyPairInfo & keyInfo(const std::string &keyRef) const;
+
 protected:
     void setAppName(const std::string &name);
+    void setInitialKeyInfos(const std::vector<KeyPairInfo> &infos);
 
     bool parseCardInfo(const std::string &name, const std::string &value);
 
+private:
+    void updateKeyInfo(const KeyPairInfo &keyPairInfo);
+
 private:
     bool mCanLearn = false;
     bool mHasNullPin = false;
@@ -101,8 +119,11 @@ private:
     std::string mCardType;
     int mCardVersion = -1;
     QString mCardHolder;
+    std::string mSigningKeyRef;
+    std::string mEncryptionKeyRef;
     std::vector<PinState> mPinStates;
     QString mErrMsg;
+    std::vector<KeyPairInfo> mKeyInfos;
 };
 } // namespace Smartcard
 } // namespace Kleopatra
